

set -ex



test -f $PREFIX/lib/libaws-c-common${SHLIB_EXT}
test -f $PREFIX/include/aws/common/config.h
test -f $PREFIX/lib/cmake/AwsCFlags.cmake
! grep Werror $PREFIX/lib/cmake/AwsCFlags.cmake
exit 0
